"""
Created on 30 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alphasense A4 electrochemical sensor
"""

from scs_core.gas.a4.a4_datum import A4Datum
from scs_core.gas.a4.a4_temp_comp import A4TempComp

from scs_core.gas.sensor import Sensor


# --------------------------------------------------------------------------------------------------------------------

class A4(Sensor):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def init(cls):
        cls.SENSORS[cls.CODE_CO] =      A4(cls.CODE_CO,     'CO A4',    'CO',   3)
        cls.SENSORS[cls.CODE_H2S] =     A4(cls.CODE_H2S,    'H2SA4',    'H2S',  3)
        cls.SENSORS[cls.CODE_NO] =      A4(cls.CODE_NO,     'NO A4',    'NO',   3)
        cls.SENSORS[cls.CODE_NO2] =     A4(cls.CODE_NO2,    'NO2A43F',  'NO2',  3)      # was NOGA4
        cls.SENSORS[cls.CODE_OX] =      A4(cls.CODE_OX,     'OXA431',   'Ox',   3)      # was OXGA4
        cls.SENSORS[cls.CODE_SO2] =     A4(cls.CODE_SO2,    'SO2A4',    'SO2',  3)

        cls.SENSORS[cls.CODE_TEST_1] =  A4(cls.CODE_TEST_1, 'TEST',     'SN1',  3)
        cls.SENSORS[cls.CODE_TEST_2] =  A4(cls.CODE_TEST_2, 'TEST',     'SN2',  3)
        cls.SENSORS[cls.CODE_TEST_3] =  A4(cls.CODE_TEST_3, 'TEST',     'SN3',  3)
        cls.SENSORS[cls.CODE_TEST_4] =  A4(cls.CODE_TEST_4, 'TEST',     'SN4',  3)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sensor_code, sensor_type, gas_name, adc_gain_index):
        """
        Constructor
        """
        Sensor.__init__(self, sensor_code, sensor_type, gas_name, adc_gain_index)

        self.__tc = A4TempComp.find(sensor_code)


    # ----------------------------------------------------------------------------------------------------------------

    def sample(self, afe, temp, sensor_index, no2_sample=None):
        we_v, ae_v = afe.sample_raw_wrk_aux(sensor_index, self.adc_gain_index)

        if self.has_no2_cross_sensitivity():
            if no2_sample is None:
                raise ValueError("A4.sample: no2_sample required, but none given.")

            return A4Datum.construct(self.calib, self.baseline, self.__tc, temp, we_v, ae_v, no2_sample.cnc)

        return A4Datum.construct(self.calib, self.baseline, self.__tc, temp, we_v, ae_v)


    def null_datum(self):
        return A4Datum(None, None)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4:{sensor_code:%s, sensor_type:%s, gas_name:%s, calib:%s, baseline:%s}" %  \
               (self.sensor_code, self.sensor_type, self.gas_name, self.calib, self.baseline)



"""
Created on 28 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python
"""

from collections import OrderedDict
from enum import Enum
from http import HTTPStatus

from scs_core.aws.data.http_response import HTTPResponse
from scs_core.data.str import Str
from scs_core.estate.configuration_check import ConfigurationCheck
from scs_core.sys.http_exception import HTTPException


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationCheckFinder(object):
    """
    classdocs
    """

    __URL = "https://p18hyi3w56.execute-api.us-west-2.amazonaws.com/default/ConfigurationCheckFinder"

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client, auth):
        self.__http_client = http_client
        self.__auth = auth


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, tag_filter, result_code, response_mode):
        request = ConfigurationCheckRequest(tag_filter, result_code, response_mode)
        headers = {'Authorization': self.__auth.email_address}

        response = self.__http_client.get(self.__URL, headers=headers, params=request.params())

        return ConfigurationCheckResponse.construct_from_jdict(response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationCheckFinder:{auth:%s}" % self.__auth


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationCheckRequest(object):
    """
    classdocs
    """

    MODE = Enum('Mode', 'FULL TAGS_ONLY')

    TAG_FILTER = 'tagFilter'
    RESULT_CODE = 'resultCode'
    RESPONSE_MODE = 'responseMode'


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        tag_filter = qsp.get(cls.TAG_FILTER)
        result_code = qsp.get(cls.RESULT_CODE)

        try:
            response_mode = cls.MODE[qsp.get(cls.RESPONSE_MODE)]
        except KeyError:
            response_mode = None

        return cls(tag_filter, result_code, response_mode)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag_filter, result_code, response_mode):
        """
        Constructor
        """
        self.__tag_filter = tag_filter                          # string
        self.__result_code = result_code                        # string
        self.__response_mode = response_mode                    # MODE enum


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.response_mode is None:
            return False

        return True


    def tags_only(self):
        return self.__response_mode == self.MODE.TAGS_ONLY


    # ----------------------------------------------------------------------------------------------------------------

    def params(self):
        params = {
            self.TAG_FILTER: self.tag_filter,
            self.RESULT_CODE: self.result_code,
            self.RESPONSE_MODE: self.response_mode.name
        }

        return params


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag_filter(self):
        return self.__tag_filter


    @property
    def result_code(self):
        return self.__result_code


    @property
    def response_mode(self):
        return self.__response_mode


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationCheckRequest:{tag_filter:%s, result_code:%s, response_mode:%s}" % \
               (self.tag_filter, self.result_code, self.response_mode)


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationCheckResponse(HTTPResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        status = HTTPStatus(jdict.get('statusCode'))

        if status != HTTPStatus.OK:
            raise HTTPException(status.value, status.phrase, status.description)

        mode = ConfigurationCheckRequest.MODE[jdict.get('mode')]

        items = []
        if jdict.get('Items'):
            for item_jdict in jdict.get('Items'):
                item = item_jdict.get('tag') if mode == ConfigurationCheckRequest.MODE.TAGS_ONLY else \
                    ConfigurationCheck.construct_from_jdict(item_jdict)
                items.append(item)

        next_url = jdict.get('next')

        return cls(status, mode, items, next_url=next_url)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, status, mode, items, next_url=None):
        """
        Constructor
        """
        super().__init__(status)

        self.__mode = mode                              # ConfigurationCheckRequest.Mode member
        self.__items = items                            # list of ConfigurationCheck or string
        self.__next_url = next_url                      # URL string


    def __len__(self):
        return len(self.items)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['statusCode'] = self.status.value

        if self.mode is not None:
            jdict['mode'] = self.mode.name

        if self.items is not None:
            jdict['Items'] = self.items
            jdict['itemCount'] = len(self.items)

        if self.next_url is not None:
            jdict['next'] = self.next_url

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def mode(self):
        return self.__mode


    @property
    def items(self):
        return self.__items


    @property
    def next_url(self):
        return self.__next_url


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationCheckResponse:{status:%s, mode:%s, items:%s, next_url:%s}" % \
               (self.status, self.mode, Str.collection(self.items), self.next_url)



"""
Created on 23 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.str import Str

from scs_core.estate.configuration import Configuration
from scs_core.sample.sample import Sample


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationSample(Sample):
    """
    classdocs
    """

    @classmethod
    def construct_from_jdict(cls, jdict, default=True):
        if not jdict:
            return None

        # Sample...
        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))
        tag = jdict.get('tag')

        try:
            val_jdict = json.loads(jdict.get('val'))
        except TypeError:
            val_jdict = jdict.get('val')

        configuration = Configuration.construct_from_jdict(val_jdict)

        return cls(tag, rec, configuration)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, configuration):
        """
        Constructor
        """
        super().__init__(tag, rec)

        self.__configuration = configuration                # Configuration


    def __eq__(self, other):
        try:
            return self.tag == other.tag and self.configuration == other.configuration

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        if self.tag < other.tag:
            return True

        if self.tag > other.tag:
            return False

        if self.rec < other.rec:
            return True

        return False


    def diff(self, other):
        return ConfigurationSample(self.tag, self.rec, self.configuration.diff(other.configuration))


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tag'] = self.tag
        jdict['rec'] = None if self.rec is None else self.rec.as_iso8601()
        jdict['val'] = self.values

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def values(self):
        return self.configuration.as_json()


    @property
    def configuration(self):
        return self.__configuration


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationSample:{tag:%s, rec:%s, configuration:%s}" % (self.tag, self.rec, self.configuration)


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationSampleHistory(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, latest_only=False, items=None):
        """
        Constructor
        """
        self.__latest_only = latest_only                            # bool
        self.__items = {} if items is None else items               # dict of tag: list of ConfigurationSample


    # ----------------------------------------------------------------------------------------------------------------

    def insert(self, sample: ConfigurationSample):
        if sample.tag not in self.__items:
            self.__items[sample.tag] = []
            self.__items[sample.tag].append(sample)
            return

        if not self.__items[sample.tag] or not self.__latest_only:
            if sample not in self.__items[sample.tag]:            # we might not be reading all the fields in the DB!
                self.__items[sample.tag].append(sample)
            return

        for item in self.__items[sample.tag]:
            if sample.rec > item.rec:
                self.__items[sample.tag] = (sample, )
                return


    # ----------------------------------------------------------------------------------------------------------------

    def diffs(self):
        diff_history = ConfigurationSampleHistory()

        for tag in self.tags():
            prev = None
            for item in self.items_for_tag(tag):
                diff_history.insert(item if prev is None else item.diff(prev))
                prev = item

        return diff_history


    def tags(self):
        return set(self.__items.keys())


    def items_for_tag(self, tag):
        if tag not in self.__items:
            return None

        return self.__items[tag]


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        items = []

        for tag in self.tags():
            items += self.items_for_tag(tag)

        return items


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationSampleHistory:{latest_only:%s, items:%s}" % \
               (self.__latest_only, Str.collection(self.__items))



"""
Created on 20 Oct 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"tag": "scs-be2-2",
 "rec": "2017-09-24T07:51:21Z",
 "val": {
  "NO2": {"weV": 0.312317, "aeV": 0.31038, "weC": -0.001, "cnc": 14.8},
  "CO": {"weV": 0.325005, "aeV": 0.254254, "weC": 0.077239, "cnc": 323.2},
  "SO2": {"weV": 0.277942, "aeV": 0.267754, "weC": 0.004136, "cnc": 27.6},
  "H2S": {"weV": 0.221816, "aeV": 0.269817, "weC": -0.006301, "cnc": 29.6},
  "pt1": {"v": 0.321411, "tmp": 21.9},
  "sht": {"hmd": 73.0, "tmp": 21.4}}}
"""

from collections import OrderedDict

from scs_core.climate.sht_datum import SHTDatum

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.str import Str

from scs_core.gas.a4.a4_datum import A4Datum
from scs_core.gas.afe.afe_datum import AFEDatum
from scs_core.gas.afe.pt1000_datum import Pt1000Datum
from scs_core.gas.pid.pid import PIDDatum
from scs_core.gas.scd30.scd30_datum import SCD30Datum

from scs_core.sample.sample import Sample


# TODO: get src from AFE / ISI datum?

# --------------------------------------------------------------------------------------------------------------------

class GasesSample(Sample):
    """
    classdocs
    """

    __NON_ELECTROCHEM_FIELDS = ['pt1', 'sht', 'CO2', 'VOC']
    __VOC_FIELDS = ['VOC']

    @classmethod
    def construct_from_jdict(cls, jdict, default=True):
        if not jdict:
            return None

        # Sample...
        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))
        tag = jdict.get('tag')
        val = jdict.get('val')
        exegeses = jdict.get('exg')

        # GasesSample...
        node = val.get('pt1')
        pt1000_datum = None if node is None else Pt1000Datum(node.get('v'), temp=node.get('tmp'))

        node = val.get('sht')
        sht_datum = SHTDatum(node.get('hmd'), node.get('tmp'))

        node = val.get('CO2')
        scd30_datum = None if node is None else SCD30Datum(node.get('cnc'), None, None)

        sns = OrderedDict()

        for field, node in val.items():
            if field not in cls.__NON_ELECTROCHEM_FIELDS:
                sns[field] = A4Datum.construct_from_jdict(node)

            if field in cls.__VOC_FIELDS:
                sns[field] = PIDDatum.construct_from_jdict(node)

        electrochem_datum = AFEDatum(pt1000_datum, *list(sns.items()))

        return cls(tag, rec, scd30_datum, electrochem_datum, sht_datum, exegeses=exegeses)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, scd30_datum, electrochem_datum, sht_datum, exegeses=None):
        """
        Constructor
        """
        super().__init__(tag, rec, exegeses=exegeses)

        self.__scd30_datum = scd30_datum                            # SCD30Datum
        self.__electrochem_datum = electrochem_datum                # AFEDatum or ISIDatum
        self.__sht_datum = sht_datum                                # SHT31Datum


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def has_invalid_value(cls):
        # TODO: implement has_invalid_value
        return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def values(self):
        jdict = OrderedDict()

        if self.scd30_datum is not None:
            jdict['CO2'] = {'cnc': self.scd30_datum.co2}

        if self.electrochem_datum is not None:
            for key, value in self.electrochem_datum.sns.items():
                jdict[key] = value.as_json()

            try:
                if self.electrochem_datum.pt1000 is not None:
                    jdict['pt1'] = self.electrochem_datum.pt1000.as_json()
            except AttributeError:
                pass

        if self.sht_datum is not None:
            jdict['sht'] = self.sht_datum.as_json()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def scd30_datum(self):
        return self.__scd30_datum


    @property
    def electrochem_datum(self):
        return self.__electrochem_datum


    @property
    def sht_datum(self):
        return self.__sht_datum


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        exegeses = Str.collection(self.exegeses)

        return "GasesSample:{tag:%s, rec:%s, exegeses:%s, scd30_datum:%s, electrochem_datum:%s, sht_datum:%s}" % \
            (self.tag, self.rec, exegeses, self.scd30_datum, self.electrochem_datum, self.sht_datum)



"""
Created on 08 Mar 2021

@author: Jade Page (jade.page@southcoastscience.com)

https://stackoverflow.com/questions/36780856/complete-scan-of-dynamodb-with-boto3
"""

from boto3.dynamodb.conditions import Key, Attr
from botocore.exceptions import ClientError

from scs_core.sys.logging import Logging


# TODO: where does sorting happen?

# --------------------------------------------------------------------------------------------------------------------

class DynamoManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, dynamo_client, dynamo_resource):
        self.__dynamo_client = dynamo_client
        self.__dynamo_resource = dynamo_resource

        self.__logger = Logging.getLogger()

    # ----------------------------------------------------------------------------------------------------------------

    def get(self, table_name, primary_key):
        # get specific item
        table = self.__dynamo_resource.Table(table_name)
        try:
            response = table.get_item(Key=primary_key)
        except ClientError as e:
            self.__logger.error(e.response['Error']['Message'])
        else:
            return response['Item'] if 'Item' in response else None

    def exists(self, table_name, primary_key_name, primary_key):
        table = self.__dynamo_resource.Table(table_name)
        try:
            response = table.query(
                KeyConditionExpression=Key(primary_key_name).eq(primary_key)
            )
        except ClientError as e:
            self.__logger.error(e.response['Error']['Message'])
        else:
            return response['Items'] if 'Items' in response else None

    def add(self, table_name, item):
        table = self.__dynamo_resource.Table(table_name)

        response = table.put_item(
            Item=item
        )
        self.__logger.info(response)

    def delete(self, table_name, item):
        table = self.__dynamo_resource.Table(table_name)
        response = None

        try:
            response = table.delete_item(
                Key=item
            )
        except ClientError as e:
            if e.response['Error']['Code'] == "ConditionalCheckFailedException":
                self.__logger.error(e.response['Error']['Message'])
            else:
                raise

        return response

    def retrieve_all(self, table_name, lek=None):
        data_dict = []
        table = self.__dynamo_resource.Table(table_name)

        if lek:
            response = table.scan(ExclusiveStartKey=lek)
        else:
            response = table.scan()

        if "Items" not in response:
            return None, None

        if response['Count'] == 0:
            return None, None

        data = response['Items']
        for item in data:
            data_dict.append(item)

        if "LastEvaluatedKey" not in response:
            return data_dict, None
        else:
            lek = response["LastEvaluatedKey"]

        while lek is not None:
            data, lek = self.retrieve_all(table_name, lek)
            try:
                data_dict += data
            except TypeError:
                lek = None

        return data_dict, lek

    def retrieve_filtered(self, table_name, filter_key, filter_value, lek=None):
        data_dict = []
        table = self.__dynamo_resource.Table(table_name)

        if lek:
            response = table.scan(
                FilterExpression=Attr(filter_key).contains(filter_value),
                ExclusiveStartKey=lek
            )
        else:
            response = table.scan(
                FilterExpression=Attr(filter_key).contains(filter_value)
            )

        if "Items" not in response:
            return None, None

        if response['Count'] == 0:
            return None, None

        data = response['Items']
        for item in data:
            data_dict.append(item)

        if "LastEvaluatedKey" not in response:
            return data_dict, None
        else:
            lek = response["LastEvaluatedKey"]

        while lek is not None:
            data, lek = self.retrieve_filtered(table_name, filter_key, filter_value, lek)
            try:
                data_dict += data
            except TypeError:
                lek = None

        return data_dict, lek

    def retrieve_all_pk(self, table_name, pk, lek=None):
        data_dict = []
        table = self.__dynamo_resource.Table(table_name)

        if lek:
            response = table.scan(
                ProjectionExpression='#pk',
                ExpressionAttributeNames={'#pk': pk},
                ExclusiveStartKey=lek
            )
        else:
            response = table.scan(
                ProjectionExpression=pk
            )

        if "Items" not in response:
            return None, None

        if response['Count'] == 0:
            return None, None

        data = response['Items']
        for item in data:
            data_dict.append(item)

        if "LastEvaluatedKey" not in response:
            return data_dict, None
        else:
            lek = response["LastEvaluatedKey"]

        while lek is not None:
            data, lek = self.retrieve_all_pk(table_name, pk, lek)
            try:
                data_dict += data
            except TypeError:
                lek = None

        return data_dict, lek

    def retrieve_filtered_pk(self, table_name, pk, tag_filter, lek=None):
        data_dict = []
        table = self.__dynamo_resource.Table(table_name)

        print(tag_filter)

        if lek:
            response = table.scan(
                FilterExpression=Attr(pk).contains(tag_filter),
                ProjectionExpression='#pk',
                ExpressionAttributeNames={'#pk': pk},
                ExclusiveStartKey=lek
            )
        else:
            response = table.scan(
                FilterExpression=Attr(pk).contains(tag_filter),
                ProjectionExpression='#pk',
                ExpressionAttributeNames={'#pk': pk}
            )

        if "Items" not in response:
            return None, None

        if response['Count'] == 0:
            return None, None

        data = response['Items']
        for item in data:
            data_dict.append(item)

        if "LastEvaluatedKey" not in response:
            return data_dict, None
        else:
            lek = response["LastEvaluatedKey"]

        while lek is not None:
            data, lek = self.retrieve_filtered_pk(table_name, pk, tag_filter, lek)
            try:
                data_dict += data
            except TypeError:
                lek = None

        return data_dict, lek

    def retrieve_double_filtered(self, table_name, first_key, first_value, second_key, second_value, lek=None):
        data_dict = []
        table = self.__dynamo_resource.Table(table_name)
        if lek:
            response = table.scan(
                FilterExpression=Attr(first_key).contains(first_value) & Attr(second_key).contains(second_value),
                ExclusiveStartKey=lek
            )
        else:
            response = table.scan(
                FilterExpression=Attr(first_key).contains(first_value) & Attr(second_key).contains(second_value)
            )

        if "Items" not in response:
            return None, None

        if response['Count'] == 0:
            return None, None

        data = response['Items']
        for item in data:
            data_dict.append(item)

        if "LastEvaluatedKey" not in response:
            return data_dict, None
        else:
            lek = response["LastEvaluatedKey"]

        while lek is not None:
            data, lek = self.retrieve_double_filtered(table_name, first_key, first_value, second_key, second_value, lek)
            try:
                data_dict += data
            except TypeError:
                lek = None

        return data_dict, lek

    def retrieve_double_filtered_pk(self, table_name, first_key, first_value, second_key, second_value, lek=None):
        data_dict = []
        table = self.__dynamo_resource.Table(table_name)
        if lek:
            response = table.scan(
                FilterExpression=Attr(first_key).contains(first_value) & Attr(second_key).contains(second_value),
                ProjectionExpression=first_key,
                ExclusiveStartKey=lek
            )
        else:
            response = table.scan(
                FilterExpression=Attr(first_key).contains(first_value) & Attr(second_key).contains(second_value),
                ProjectionExpression=first_key
            )

        if "Items" not in response:
            return None, None

        if response['Count'] == 0:
            return None, None

        data = response['Items']
        for item in data:
            data_dict.append(item)

        if "LastEvaluatedKey" not in response:
            return data_dict, None
        else:
            lek = response["LastEvaluatedKey"]

        while lek is not None:
            data, lek = self.retrieve_double_filtered_pk(table_name, first_key, first_value, second_key, second_value,
                                                         lek)
            try:
                data_dict += data
            except TypeError:
                lek = None

        return data_dict, lek

    def filter_on_second_value(self, table_name, pk, second_key, second_value, lek=None):
        data_dict = []
        table = self.__dynamo_resource.Table(table_name)
        if lek:
            response = table.scan(
                FilterExpression=Attr(second_key).contains(second_value),
                ProjectionExpression=pk,
                ExclusiveStartKey=lek
            )
        else:
            response = table.scan(
                FilterExpression=Attr(second_key).contains(second_value),
                ProjectionExpression=pk
            )

        if "Items" not in response:
            return None, None

        if response['Count'] == 0:
            return None, None

        data = response['Items']
        for item in data:
            data_dict.append(item)

        if "LastEvaluatedKey" not in response:
            return data_dict, None
        else:
            lek = response["LastEvaluatedKey"]

        while lek is not None:
            data, lek = self.filter_on_second_value(table_name, pk, second_key, second_value, lek)
            try:
                data_dict += data
            except TypeError:
                lek = None

        return data_dict, lek



"""
Created on 13 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/42568262/how-to-encrypt-text-with-a-password-in-python
"""

import json
import os
import time

from abc import abstractmethod


# --------------------------------------------------------------------------------------------------------------------

class JSONify(json.JSONEncoder):
    """
    classdocs
    """

    @staticmethod
    def dumps(obj, skipkeys=False, ensure_ascii=False, check_circular=True,
              allow_nan=True, cls=None, indent=None, separators=None,
              default=None, sort_keys=False, **kw):

        handler = JSONify if cls is None else cls

        return json.dumps(obj, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
                          allow_nan=allow_nan, cls=handler, indent=indent, separators=separators,
                          default=default, sort_keys=sort_keys, **kw)


    # ----------------------------------------------------------------------------------------------------------------

    def default(self, obj):
        if isinstance(obj, JSONable):
            return obj.as_json()

        return json.JSONEncoder.default(self, obj)


# --------------------------------------------------------------------------------------------------------------------

class JSONable(object):
    """
    classdocs
    """

    _INDENT = 4

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def loads(cls, jstr):
        try:
            return json.loads(jstr)
        except json.decoder.JSONDecodeError:
            raise ValueError(jstr.strip())


    # ----------------------------------------------------------------------------------------------------------------

    def as_list(self, jlist):
        del jlist[:]                                    # empty the list

        for key, value in self.as_json().items():
            try:
                value = value.as_json()
            except AttributeError:
                pass

            jlist.append((key, value))                  # append the key-value pairs of the dictionary


    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def as_json(self, *args, **kwargs):
        pass


# --------------------------------------------------------------------------------------------------------------------

class JSONReport(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def load(cls, filename, default=True):
        if filename is None:
            return None

        if not os.path.isfile(filename):
            return cls.construct_from_jdict(None, default=default)

        with open(filename, 'r') as f:
            jstr = f.read()

        return cls.construct_from_jdict(cls.loads(jstr), default=default)


    @classmethod
    def delete(cls, filename):
        if filename is None:
            return

        os.remove(filename)


    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict, default=True):
        return JSONReport()


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, filename):
        if filename is None:
            return

        # data...
        jstr = JSONify.dumps(self, indent=self._INDENT)

        # file...
        tmp_filename = '.'.join((filename, str(int(time.time()))))

        try:
            with open(tmp_filename, 'w') as f:
                f.write(jstr + '\n')

        except FileNotFoundError:           # the containing directory does not exist (yet)
            return False

        # atomic operation...
        os.rename(tmp_filename, filename)

        return True


# --------------------------------------------------------------------------------------------------------------------

class AbstractPersistentJSONable(JSONable):
    """
    classdocs
    """

    __AWS_DIR =             "aws"                               # hard-coded rel path
    __CONF_DIR =            "conf"                              # hard-coded rel path
    __HUE_DIR =             "hue"                               # hard-coded rel path
    __OSIO_DIR =            "osio"                              # hard-coded rel path

    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def save(self, manager):
        pass

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def aws_dir(cls):
        return cls.__AWS_DIR


    @classmethod
    def conf_dir(cls):
        return cls.__CONF_DIR


    @classmethod
    def hue_dir(cls):
        return cls.__HUE_DIR


    @classmethod
    def osio_dir(cls):
        return cls.__OSIO_DIR


# --------------------------------------------------------------------------------------------------------------------

class PersistentJSONable(AbstractPersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def exists(cls, manager):
        try:
            dirname, filename = cls.persistence_location()
        except NotImplementedError:
            return False

        return manager.exists(dirname, filename)


    @classmethod
    def load(cls, manager, encryption_key=None, default=True):
        try:
            dirname, filename = cls.persistence_location()
        except NotImplementedError:
            return None

        if not manager.exists(dirname, filename):
            return cls.construct_from_jdict(None, default=default)

        jstr = manager.load(dirname, filename, encryption_key=encryption_key)

        return cls.construct_from_jdict(cls.loads(jstr), default=default)


    @classmethod
    def delete(cls, manager):
        try:
            dirname, filename = cls.persistence_location()
            manager.remove(dirname, filename)

        except NotImplementedError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict, default=True):
        return PersistentJSONable()


    @classmethod
    @abstractmethod
    def persistence_location(cls):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager, encryption_key=None):
        dirname, filename = self.persistence_location()
        jstr = JSONify.dumps(self, indent=self._INDENT)

        manager.save(jstr, dirname, filename, encryption_key=encryption_key)


# --------------------------------------------------------------------------------------------------------------------

class MultiPersistentJSONable(AbstractPersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def exists(cls, manager, name=None):
        try:
            dirname, filename = cls.persistence_location(name)
        except NotImplementedError:
            return False

        return manager.exists(dirname, filename)


    @classmethod
    def load(cls, manager, name=None, encryption_key=None, default=True):
        try:
            dirname, filename = cls.persistence_location(name)
        except NotImplementedError:
            return None

        if not manager.exists(dirname, filename):
            return cls.construct_from_jdict(None, name=name, default=default)

        jstr = manager.load(dirname, filename, encryption_key=encryption_key)

        return cls.construct_from_jdict(cls.loads(jstr), name=name, default=default)


    @classmethod
    def delete(cls, manager, name=None):
        try:
            dirname, filename = cls.persistence_location(name)
            manager.remove(dirname, filename)

        except NotImplementedError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict, name=None, default=True):
        return PersistentJSONable()


    @classmethod
    @abstractmethod
    def persistence_location(cls, name):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name):
        self.__name = name                                          # string


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager, encryption_key=None):
        dirname, filename = self.persistence_location(self.name)
        jstr = JSONify.dumps(self, indent=self._INDENT)

        manager.save(jstr, dirname, filename, encryption_key=encryption_key)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        return self.__name


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MultiPersistentJSONable:{name:%s}" % self.name



"""
Created on 08 Feb 2021

@author: Jade Page (jade.page@southcoastscience.com)
"""

import json
import os
import re

from collections import OrderedDict

from scs_core.aws.greengrass.aws_group import AWSGroup

from scs_core.data.path_dict import PathDict

from scs_core.estate.configuration import Configuration

from scs_core.sys.logging import Logging
from scs_core.sys.system_id import SystemID


# --------------------------------------------------------------------------------------------------------------------

class AWSGroupCloner(object):
    """
    classdocs
    """

    __CWD = os.path.dirname(os.path.realpath(__file__))
    __V1 = __CWD + "/v1"

    __GAS_FUNCTION = "GasInference"
    __PARTS_FUNCTION = "PMxInference"

    __GAS_RESOURCE = "ML-NO2"
    __PARTS_RESOURCE = ["ML-PM1", "ML-PM10", "ML-PM2P5"]

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, origin_group_name, dest_group_name, client, gas, parts):
        self.__gas = gas
        self.__parts = parts

        self.__origin_group_name = origin_group_name
        self.__origin_group_number = None
        self.__dest_group_name = dest_group_name
        self.__dest_group_number = None
        self.__origin_owner = None
        self.__dest_owner = None

        self.__logger = Logging.getLogger()

        self.__client = client
        self.__aws_group_info = AWSGroup(self.__origin_group_name, self.__client)

        self.__aws_info = PathDict()

        self.__arns = None
        self.__origin_function_info = None
        self.__origin_resource_info = None

        self.__dest_function_info = None
        self.__dest_resource_info = None
        self.__dest_loc_path = None
        self.__dest_dev_path = None
        self.__dest_group_id = None


    # ----------------------------------------------------------------------------------------------------------------

    def retrieve_group_info(self):
        aws_json_reader = AWSGroup(self.__origin_group_name, self.__client)
        aws_json_reader.get_group_info_from_name()
        aws_json_reader.get_group_arns()
        aws_json_reader.get_function_definition()
        aws_json_reader.get_logger_definition()
        aws_json_reader.get_resource_definition()
        aws_json_reader.get_subscription_definition_version()

        origin_info = aws_json_reader.return_verbose_info()

        function_info = origin_info.node("Function Definition Response")
        self.__origin_function_info = function_info["Definition"]

        resource_info = origin_info.node("Resource Definition Response")
        self.__origin_resource_info = resource_info["Definition"]

        aws_json_reader = AWSGroup(self.__dest_group_name, self.__client)
        group_info = aws_json_reader.get_group_info_from_name()
        group_id = group_info.node("GroupID")
        self.__dest_group_id = group_id
        aws_json_reader.get_group_arns()
        self.__aws_info.append("CoreDefinitionARN", aws_json_reader.retrieve_node("CoreDefinitionVersionArn"))


    def update_function_info(self):
        new_dict = []

        for item in self.__origin_function_info["Functions"]:
            resource = json.dumps(item)

            # There's probably a more elegant way to do this:
            if self.__GAS_FUNCTION in resource:
                if not self.__gas:
                    continue

            if self.__PARTS_FUNCTION in resource:
                if not self.__parts:
                    continue

            n_resource = resource.replace(self.__origin_group_number, self.__dest_group_number)
            new_dict.append(json.loads(n_resource))

        self.__dest_function_info = new_dict


    def update_resource_info(self):
        new_dict = []

        for item in self.__origin_resource_info["Resources"]:

            if self.__GAS_RESOURCE in item["Name"]:
                if not self.__gas:
                    continue

            if item["Name"] in self.__PARTS_RESOURCE:
                if not self.__parts:
                    continue

            resource = json.dumps(item)
            no_resource = resource.replace(self.__origin_group_number, self.__dest_group_number)
            n_resource = no_resource.replace(self.__origin_owner, self.__dest_owner)
            new_dict.append(json.loads(n_resource))

        self.__dest_resource_info = new_dict


    def validate_names(self):
        reg = re.compile("scs-(\\w+)-(\\d{3})-group")
        is_valid = reg.fullmatch(self.__origin_group_name)
        if not is_valid:
            return False

        is_valid = reg.fullmatch(self.__dest_group_name)
        if not is_valid:
            return False

        return True


    def create_resources_version(self):
        jdict = OrderedDict()

        resources = self.__dest_resource_info
        resource_container_name = "Resources-" + self.__dest_group_number
        jdict['Name'] = resource_container_name

        jdict_init = OrderedDict()
        jdict_init['Resources'] = resources

        jdict['InitialVersion'] = jdict_init

        # Send request
        self.__logger.info("Creating resource definition")
        self.__logger.debug(jdict)

        response = self.__client.create_resource_definition(InitialVersion=jdict["InitialVersion"])
        self.__aws_info.append("NewResourceARN", response["LatestVersionArn"])

        self.__logger.debug(response)


    def create_function_version(self):
        jdict = OrderedDict()

        functions = self.__dest_function_info

        jdict_init = OrderedDict()
        jdict_init['Functions'] = functions

        jdict['InitialVersion'] = jdict_init

        self.__logger.info("Creating function definition")
        self.__logger.debug(jdict)

        # Create request
        response = self.__client.create_function_definition(InitialVersion=jdict["InitialVersion"])
        self.__aws_info.append("NewFunctionARN", response["LatestVersionArn"])

        self.__logger.debug(response)


    def get_project_locations(self, host):
        system_id = SystemID.load(host)

        if system_id is None:
            self.__logger.error('SystemID not available.')
            return False

        configuration = Configuration.load(host)

        self.__dest_loc_path = configuration.aws_project.location_path
        self.__dest_dev_path = configuration.aws_project.device_path

        return True


    def create_subscription_version(self):
        # Get template JSON
        j_file = os.path.join(self.__V1, 'gg_subscriptions.json')
        with open(j_file) as f:
            data = f.read()

        # Edit for device
        group_number = self.__dest_group_number
        device_path = self.__dest_dev_path
        loc_path = self.__dest_loc_path

        s_data = json.loads(data)
        s_data["InitialVersion"]["Subscriptions"][0]["Id"] = (group_number + "-particulates-subscription")
        s_data["InitialVersion"]["Subscriptions"][1]["Id"] = (group_number + "-control-from-cloud-subscription")
        s_data["InitialVersion"]["Subscriptions"][2]["Id"] = (group_number + "-climate-subscription")
        s_data["InitialVersion"]["Subscriptions"][3]["Id"] = (group_number + "-status-subscription")
        s_data["InitialVersion"]["Subscriptions"][4]["Id"] = (group_number + "-control-to-cloud-subscription")
        s_data["InitialVersion"]["Subscriptions"][5]["Id"] = (group_number + "-gases-subscription")

        s_data["InitialVersion"]["Subscriptions"][0]["Subject"] = (loc_path + "/particulates")
        s_data["InitialVersion"]["Subscriptions"][1]["Subject"] = (device_path + "/control")
        s_data["InitialVersion"]["Subscriptions"][2]["Subject"] = (loc_path + "/climate")
        s_data["InitialVersion"]["Subscriptions"][3]["Subject"] = (device_path + "/status")
        s_data["InitialVersion"]["Subscriptions"][4]["Subject"] = (device_path + "/control")
        s_data["InitialVersion"]["Subscriptions"][5]["Subject"] = (loc_path + "/gases")

        self.__logger.info("Creating sub definition")
        self.__logger.debug(s_data)

        # Send request
        response = self.__client.create_subscription_definition(InitialVersion=s_data["InitialVersion"])
        self.__aws_info.append("NewSubscriptionARN", response["LatestVersionArn"])

        self.__logger.debug(response)


    def create_aws_group_definition(self):
        self.__logger.info("Creating group definition")

        response = self.__client.create_group_version(
            CoreDefinitionVersionArn=self.__aws_info.node("CoreDefinitionARN"),
            FunctionDefinitionVersionArn=self.__aws_info.node("NewFunctionARN"),
            GroupId=self.__dest_group_id,
            ResourceDefinitionVersionArn=self.__aws_info.node("NewResourceARN"),
            SubscriptionDefinitionVersionArn=self.__aws_info.node("NewSubscriptionARN"),
        )

        self.__logger.info(response)


    def run(self, host):
        temp = self.__origin_group_name.split("-")
        self.__origin_group_number = temp[2]
        self.__origin_owner = "pi" if temp[1] == "rpi" else "scs"

        temp = self.__dest_group_name.split("-")
        self.__dest_group_number = temp[2]
        self.__dest_owner = "pi" if temp[1] == "rpi" else "scs"

        self.retrieve_group_info()

        self.update_function_info()
        self.update_resource_info()

        self.create_resources_version()
        self.create_function_version()

        if not self.get_project_locations(host):
            return False

        self.create_subscription_version()

        self.create_aws_group_definition()

        return True  # Succeeded probably



"""
Created on 1 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"calibrated-on": "2019-02-02T11:34:16Z", "offset": 50, "env": {"hmd": 66.0, "tmp": 11.0, "pA": 99.0}}
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class SensorBaseline(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return SensorBaseline(None, 0, None)

        if 'calibrated_on' in jdict:                            # TODO: deprecated
            date = Datum.date(jdict.get('calibrated_on'))
            calibrated_on = LocalizedDatetime.construct_from_date(date)

        else:
            calibrated_on = Datum.datetime(jdict.get('calibrated-on'))

        offset = jdict.get('offset')
        environment = BaselineEnvironment.construct_from_jdict(jdict.get('env'))

        return SensorBaseline(calibrated_on, offset, environment)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, calibrated_on, offset, environment):
        """
        Constructor
        """
        self.__calibrated_on = calibrated_on            # LocalizedDatetime

        self.__offset = int(offset)                     # int                       ppb
        self.__environment = environment                # BaselineEnvironment


    def __eq__(self, other):
        try:
            return self.calibrated_on == other.calibrated_on and self.offset == other.offset and \
                   self.environment == other.environment

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['calibrated-on'] = None if self.calibrated_on is None else self.calibrated_on.as_iso8601(False)

        jdict['offset'] = self.offset
        jdict['env'] = self.environment

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def calibrated_on(self):
        return self.__calibrated_on


    @property
    def offset(self):
        return self.__offset


    @property
    def environment(self):
        return self.__environment


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SensorBaseline:{calibrated_on:%s, offset:%s, environment:%s}" % \
               (self.calibrated_on, self.offset, self.environment)


# --------------------------------------------------------------------------------------------------------------------

class BaselineEnvironment(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        humid = jdict.get('hmd')
        temp = jdict.get('tmp')
        press = jdict.get('pA')

        return BaselineEnvironment(humid, temp, press)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, humid, temp, press):
        """
        Constructor
        """
        self.__humid = Datum.float(humid, 1)            # float                 %
        self.__temp = Datum.float(temp, 1)              # float                 °C
        self.__press = Datum.float(press, 1)            # float                 kPa


    def __eq__(self, other):
        try:
            return self.humid == other.humid and self.temp == other.temp and self.press == other.press

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['hmd'] = self.humid
        jdict['tmp'] = self.temp
        jdict['pA'] = self.press

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def humid(self):
        return self.__humid


    @property
    def temp(self):
        return self.__temp


    @property
    def press(self):
        return self.__press


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "BaselineEnvironment:{humid:%s, temp:%s, press:%s}" % (self.humid, self.temp, self.press)



"""
Created on 11 Jan 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json

from scs_core.control.control_receipt import ControlReceipt


# --------------------------------------------------------------------------------------------------------------------

class ControlHandler(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, host_tag, device_tag):
        """
        Constructor
        """
        self.__host_tag = host_tag                      # string
        self.__device_tag = device_tag                  # string

        self.__outgoing_pub = None                      # ControlDatum
        self.__receipt = None                           # ControlReceipt


    # ----------------------------------------------------------------------------------------------------------------

    def set_outgoing(self, outgoing_pub):
        self.__outgoing_pub = outgoing_pub
        self.__receipt = None


    def handle(self, _client, _userdata, message):
        payload = json.loads(message.payload.decode())

        try:
            receipt = ControlReceipt.construct_from_jdict(payload)
        except TypeError:
            return

        if receipt.tag != self.__device_tag:
            return

        if receipt.attn is not None and receipt.attn != self.__host_tag:
            return

        if receipt.omd != self.__outgoing_pub.payload.digest:
            return

        self.__receipt = receipt


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def receipt(self):
        return self.__receipt


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ControlHandler:{host_tag:%s, device_tag:%s, outgoing_pub:%s, receipt:%s}" %  \
               (self.__host_tag, self.__device_tag, self.__outgoing_pub, self.receipt)



"""
Created on 19 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Based-on code
https://github.com/hackscribble/microbit-MPL115A1-barometer/blob/master/microbit-MPL115A1-barometer.py
https://gist.github.com/cubapp/23dd4e91814a995b8ff06f406679abcf
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable

from scs_core.sample.sample import Sample


# --------------------------------------------------------------------------------------------------------------------

class MPL115A2Datum(JSONable):
    """
    NXP MPL115A2 digital barometer - data interpretation
    """

    # ----------------------------------------------------------------------------------------------------------------

    __PRESSURE_CONV = (115.0 - 50.0) / 1023.0

    __DEFAULT_C25 = 472                                 # T adc counts at 25 °C
    __COUNTS_PER_DEGREE = -5.35                         # T adc counts per degree centigrade


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __actual_press(cls, p_comp):
        return p_comp * cls.__PRESSURE_CONV + 50.0


    @classmethod
    def __sl_press(cls, actual_press, temp, altitude):
        if temp is None or altitude is None:
            return None

        return (actual_press * 9.80665 * altitude) / (287 * (273 + temp + (altitude / 400))) + actual_press


    @classmethod
    def __temp(cls, c25, t_adc):
        if c25 is None:
            return None

        return (t_adc - c25) / cls.__COUNTS_PER_DEGREE + 25.0


    @classmethod
    def __c25(cls, ref_temp, t_adc):
        return round(t_adc - cls.__COUNTS_PER_DEGREE * (ref_temp - 25.0))


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, c25, p_comp, t_adc, altitude, include_temp=True):
        if p_comp is None or t_adc is None:
            return None

        temp = cls.__temp(c25, t_adc)

        actual_press = cls.__actual_press(p_comp)
        sl_press = cls.__sl_press(actual_press, temp, altitude)

        reported_temp = temp if include_temp else None

        return cls(actual_press, sl_press, t_adc, reported_temp)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, actual_press, sl_press, t_adc, temp):
        """
        Constructor
        """
        self.__actual_press = Datum.float(actual_press, 1)      # kPa
        self.__sl_press = Datum.float(sl_press, 1)              # kPa

        self.__t_adc = Datum.int(t_adc)                         # T adc count
        self.__temp = Datum.float(temp, 1)                      # °C


    # ----------------------------------------------------------------------------------------------------------------

    def as_sample(self, tag, rec):                          # TODO: remove as_sample(..)
        return Sample(tag, rec, values=self.as_json())


    def as_json(self):
        jdict = OrderedDict()

        jdict['pA'] = self.actual_press

        if self.sl_press is not None:
            jdict['p0'] = self.sl_press

        if self.temp is not None:
            jdict['tmp'] = self.temp

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def c25(self, ref_temp):
        return MPL115A2Datum.__c25(ref_temp, self.__t_adc)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def actual_press(self):
        return self.__actual_press


    @property
    def sl_press(self):
        return self.__sl_press


    @property
    def temp(self):
        return self.__temp


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MPL115A2Datum:{actual_press:%s, sl_press:%s, t_adc:%s, temp:%s}" % \
               (self.actual_press, self.sl_press, self.__t_adc, self.temp)



"""
Created on 22 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

an ML model group configuration for gases inference

example JSON:
{"uds-path": "pipes/lambda-model-gas-s1.uds", "model-interface": "s1"}
"""

from scs_core.gas.afe_calib import AFECalib

from scs_core.model.gas.gas_inference_client import GasInferenceClient
from scs_core.model.model_conf import ModelConf

from scs_core.sync.schedule import ScheduleItem


# --------------------------------------------------------------------------------------------------------------------

class GasModelConf(ModelConf):
    """
    classdocs
    """

    __FILENAME = "gas_model_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    __INTERFACES = ['s1', 'vB']

    @classmethod
    def interfaces(cls):
        return cls.__INTERFACES


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_path, model_interface):
        """
        Constructor
        """
        super().__init__(uds_path, model_interface)


    # ----------------------------------------------------------------------------------------------------------------

    def client(self, host, socket, gas_schedule_item: ScheduleItem, afe_calib: AFECalib) -> GasInferenceClient:
        if self.model_interface == 's1':
            from scs_core.model.gas.s1.s1_gas_inference_client import S1GasInferenceClient

            return S1GasInferenceClient.construct(socket, self.abs_uds_path(host), gas_schedule_item, afe_calib)

        if self.model_interface == 'vB':
            from scs_core.model.gas.vB.vb_gas_inference_client import VBGasInferenceClient

            return VBGasInferenceClient.construct(socket, self.abs_uds_path(host), gas_schedule_item)

        raise ValueError(self.model_interface)



"""
Created on 27 Jan 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json
import socket

from collections import OrderedDict

from scs_core.aws.client.api_auth import APIAuth
from scs_core.aws.client.client_auth import ClientAuth
from scs_core.aws.config.project import Project
from scs_core.aws.greengrass.aws_group_configuration import AWSGroupConfiguration
from scs_core.aws.greengrass.aws_identity import AWSIdentity

from scs_core.climate.mpl115a2_calib import MPL115A2Calib
from scs_core.climate.mpl115a2_conf import MPL115A2Conf
from scs_core.climate.sht_conf import SHTConf

from scs_core.comms.mqtt_conf import MQTTConf

from scs_core.csv.csv_logger_conf import CSVLoggerConf

from scs_core.data.json import JSONable

from scs_core.display.display_conf import DisplayConf

from scs_core.estate.package_version import PackageVersions

from scs_core.gas.afe_baseline import AFEBaseline
from scs_core.gas.afe_id import AFEId
from scs_core.gas.afe.pt1000_calib import Pt1000Calib
from scs_core.gas.ndir.ndir_conf import NDIRConf
from scs_core.gas.scd30.scd30_conf import SCD30Conf
from scs_core.gas.scd30.scd30_baseline import SCD30Baseline

from scs_core.gps.gps_conf import GPSConf

from scs_core.interface.interface_conf import InterfaceConf

from scs_core.location.timezone_conf import TimezoneConf

from scs_core.model.gas.gas_baseline import GasBaseline
from scs_core.model.gas.gas_model_conf import GasModelConf
from scs_core.model.pmx.pmx_model_conf import PMxModelConf

from scs_core.particulate.opc_conf import OPCConf

from scs_core.psu.psu_conf import PSUConf
from scs_core.psu.psu_version import PSUVersion

from scs_core.sync.schedule import Schedule

from scs_core.sys.modem import Modem, SIM
from scs_core.sys.network import Networks
from scs_core.sys.shared_secret import SharedSecret
from scs_core.sys.system_id import SystemID


# --------------------------------------------------------------------------------------------------------------------

class Configuration(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jstr(cls, jstr):
        try:
            return cls.construct_from_jdict(json.loads(jstr))
        except json.decoder.JSONDecodeError:
            return None


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        hostname = jdict.get('hostname')
        packs = PackageVersions.construct_from_jdict(jdict.get('packs'))

        afe_baseline = AFEBaseline.construct_from_jdict(jdict.get('afe-baseline'), default=None)
        afe_id = AFEId.construct_from_jdict(jdict.get('afe-id'), default=None)
        aws_api_auth = APIAuth.construct_from_jdict(jdict.get('aws-api-auth'), default=None)
        aws_client_auth = ClientAuth.construct_from_jdict(jdict.get('aws-client-auth'), default=None)
        aws_group_config = AWSGroupConfiguration.construct_from_jdict(jdict.get('aws-group-config'), default=None)
        aws_project = Project.construct_from_jdict(jdict.get('aws-project'), default=None)
        csv_logger_conf = CSVLoggerConf.construct_from_jdict(jdict.get('csv-logger-conf'), default=None)
        display_conf = DisplayConf.construct_from_jdict(jdict.get('display-conf'), default=None)
        gas_baseline = GasBaseline.construct_from_jdict(jdict.get('gas-baseline'), default=None)
        gas_model_conf = GasModelConf.construct_from_jdict(jdict.get('gas-model-conf'), default=None)
        gps_conf = GPSConf.construct_from_jdict(jdict.get('gps-conf'), default=None)
        greengrass_identity = AWSIdentity.construct_from_jdict(jdict.get('greengrass-identity'), default=None)
        interface_conf = InterfaceConf.construct_from_jdict(jdict.get('interface-conf'), default=None)
        mpl115a2_calib = MPL115A2Calib.construct_from_jdict(jdict.get('mpl115a2-calib'), default=None)
        mpl115a2_conf = MPL115A2Conf.construct_from_jdict(jdict.get('mpl115a2-conf'), default=None)
        mqtt_conf = MQTTConf.construct_from_jdict(jdict.get('mqtt-conf'), default=None)
        ndir_conf = NDIRConf.construct_from_jdict(jdict.get('ndir-conf'), default=None)
        opc_conf = OPCConf.construct_from_jdict(jdict.get('opc-conf'), default=None)
        pmx_model_conf = PMxModelConf.construct_from_jdict(jdict.get('pmx-model-conf'), default=None)
        psu_conf = PSUConf.construct_from_jdict(jdict.get('psu-conf'), default=None)
        psu_version = PSUVersion.construct_from_jdict(jdict.get('psu-version'))
        pt1000_calib = Pt1000Calib.construct_from_jdict(jdict.get('pt1000-calib'), default=None)
        scd30_baseline = SCD30Baseline.construct_from_jdict(jdict.get('scd30-baseline'), default=None)
        scd30_conf = SCD30Conf.construct_from_jdict(jdict.get('scd30-conf'), default=None)
        schedule = Schedule.construct_from_jdict(jdict.get('schedule'), default=None)
        shared_secret = SharedSecret.construct_from_jdict(jdict.get('shared-secret'), default=None)
        sht_conf = SHTConf.construct_from_jdict(jdict.get('sht-conf'), default=None)
        networks = Networks.construct_from_jdict(jdict.get('networks'))
        modem = Modem.construct_from_jdict(jdict.get('modem'))
        sim = SIM.construct_from_jdict(jdict.get('sim'))
        system_id = SystemID.construct_from_jdict(jdict.get('system-id'), default=None)
        timezone_conf = TimezoneConf.construct_from_jdict(jdict.get('timezone-conf'), default=None)

        return cls(hostname, packs, afe_baseline, afe_id, aws_api_auth,
                   aws_client_auth, aws_group_config, aws_project, csv_logger_conf, display_conf,
                   gas_baseline, gas_model_conf, gps_conf, greengrass_identity, interface_conf,
                   mpl115a2_calib, mpl115a2_conf, mqtt_conf, ndir_conf, opc_conf,
                   pmx_model_conf, psu_conf, psu_version, pt1000_calib, scd30_baseline,
                   scd30_conf, schedule, shared_secret, sht_conf, networks, modem,
                   sim, system_id, timezone_conf)


    @classmethod
    def load(cls, manager, psu=None):
        hostname = socket.gethostname()
        packs = PackageVersions.construct_from_installation(manager.scs_path())

        afe_baseline = AFEBaseline.load(manager, default=None)
        afe_id = AFEId.load(manager, default=None)
        aws_api_auth = APIAuth.load(manager, default=None)
        aws_client_auth = ClientAuth.load(manager, default=None)
        aws_group_config = AWSGroupConfiguration.load(manager, default=None)
        aws_project = Project.load(manager, default=None)
        csv_logger_conf = CSVLoggerConf.load(manager, default=None)
        display_conf = DisplayConf.load(manager, default=None)
        gas_baseline = GasBaseline.load(manager, default=None)
        gas_model_conf = GasModelConf.load(manager, default=None)
        gps_conf = GPSConf.load(manager, default=None)
        greengrass_identity = AWSIdentity.load(manager, default=None)
        interface_conf = InterfaceConf.load(manager, default=None)
        mpl115a2_calib = MPL115A2Calib.load(manager, default=None)
        mpl115a2_conf = MPL115A2Conf.load(manager, default=None)
        mqtt_conf = MQTTConf.load(manager, default=None)
        ndir_conf = NDIRConf.load(manager, default=None)
        opc_conf = OPCConf.load(manager, default=None)
        pmx_model_conf = PMxModelConf.load(manager, default=None)
        psu_conf = PSUConf.load(manager, default=None)
        psu_version = None if psu is None else psu.version()
        pt1000_calib = Pt1000Calib.load(manager, default=None)
        scd30_baseline = SCD30Baseline.load(manager, default=None)
        scd30_conf = SCD30Conf.load(manager, default=None)
        schedule = Schedule.load(manager, default=None)
        shared_secret = SharedSecret.load(manager, default=None)
        sht_conf = SHTConf.load(manager, default=None)
        networks = manager.networks()
        modem = manager.modem()
        sim = manager.sim()
        system_id = SystemID.load(manager, default=None)
        timezone_conf = TimezoneConf.load(manager, default=None)

        return cls(hostname, packs, afe_baseline, afe_id, aws_api_auth,
                   aws_client_auth, aws_group_config, aws_project, csv_logger_conf, display_conf,
                   gas_baseline, gas_model_conf, gps_conf, greengrass_identity, interface_conf,
                   mpl115a2_calib, mpl115a2_conf, mqtt_conf, ndir_conf, opc_conf,
                   pmx_model_conf, psu_conf, psu_version, pt1000_calib, scd30_baseline,
                   scd30_conf, schedule, shared_secret, sht_conf, networks, modem,
                   sim, system_id, timezone_conf)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, hostname, packs, afe_baseline, afe_id, aws_api_auth,
                 aws_client_auth, aws_group_config, aws_project, csv_logger_conf, display_conf,
                 gas_baseline, gas_model_conf, gps_conf, greengrass_identity, interface_conf,
                 mpl115a2_calib, mpl115a2_conf, mqtt_conf, ndir_conf, opc_conf,
                 pmx_model_conf, psu_conf, psu_version, pt1000_calib, scd30_baseline,
                 scd30_conf, schedule, shared_secret, sht_conf, networks, modem,
                 sim, system_id, timezone_conf):
        """
        Constructor
        """

        self.__hostname = hostname                                  # string
        self.__packs = packs                                        # PackageVersions

        self.__afe_baseline = afe_baseline                          # AFEBaseline
        self.__afe_id = afe_id                                      # AFEId
        self.__aws_api_auth = aws_api_auth                          # APIAuth
        self.__aws_client_auth = aws_client_auth                    # ClientAuth
        self.__aws_group_config = aws_group_config                  # AWSGroupConfiguration
        self.__aws_project = aws_project                            # Project
        self.__csv_logger_conf = csv_logger_conf                    # CSVLoggerConf
        self.__display_conf = display_conf                          # DisplayConf
        self.__gas_baseline = gas_baseline                          # GasBaseline
        self.__gas_model_conf = gas_model_conf                      # GasModelConf
        self.__gps_conf = gps_conf                                  # GPSConf
        self.__greengrass_identity = greengrass_identity            # AWSIdentity
        self.__interface_conf = interface_conf                      # InterfaceConf
        self.__mpl115a2_calib = mpl115a2_calib                      # MPL115A2Calib
        self.__mpl115a2_conf = mpl115a2_conf                        # MPL115A2Conf
        self.__mqtt_conf = mqtt_conf                                # MQTTConf
        self.__ndir_conf = ndir_conf                                # NDIRConf
        self.__opc_conf = opc_conf                                  # OPCConf
        self.__pmx_model_conf = pmx_model_conf                      # PMxModelConf
        self.__psu_conf = psu_conf                                  # PSUConf
        self.__psu_version = psu_version                            # PSUVersion
        self.__pt1000_calib = pt1000_calib                          # Pt1000Calib
        self.__scd30_baseline = scd30_baseline                      # SCD30Baseline
        self.__scd30_conf = scd30_conf                              # SCD30Conf
        self.__schedule = schedule                                  # Schedule
        self.__shared_secret = shared_secret                        # SharedSecret
        self.__sht_conf = sht_conf                                  # SHTConf
        self.__networks = networks                                  # Networks
        self.__modem = modem                                        # Modem
        self.__sim = sim                                            # SIM
        self.__system_id = system_id                                # SystemID
        self.__timezone_conf = timezone_conf                        # TimezoneConf


    def __eq__(self, other):
        try:
            return self.hostname == other.hostname and self.packs == other.packs and \
                   self.afe_baseline == other.afe_baseline and self.afe_id == other.afe_id and \
                   self.aws_api_auth == other.aws_api_auth and self.aws_client_auth == other.aws_client_auth and \
                   self.aws_group_config == other.aws_group_config and self.aws_project == other.aws_project and \
                   self.csv_logger_conf == other.csv_logger_conf and self.display_conf == other.display_conf and \
                   self.gas_baseline == other.gas_baseline and self.gas_model_conf == other.gas_model_conf and \
                   self.gps_conf == other.gps_conf and self.greengrass_identity == other.greengrass_identity and \
                   self.interface_conf == other.interface_conf and self.mpl115a2_calib == other.mpl115a2_calib and \
                   self.mpl115a2_conf == other.mpl115a2_conf and self.mqtt_conf == other.mqtt_conf and \
                   self.ndir_conf == other.ndir_conf and self.opc_conf == other.opc_conf and \
                   self.pmx_model_conf == other.pmx_model_conf and self.psu_conf == other.psu_conf and \
                   self.psu_version == other.psu_version and self.pt1000_calib == other.pt1000_calib and \
                   self.scd30_baseline == other.scd30_baseline and self.scd30_conf == other.scd30_conf and \
                   self.schedule == other.schedule and self.shared_secret == other.shared_secret and \
                   self.sht_conf == other.sht_conf and self.networks == other.networks and \
                   self.modem == other.modem and self.sim == other.sim and \
                   self.system_id == other.system_id and self.timezone_conf == other.timezone_conf

        except (TypeError, AttributeError):
            return False


    def diff(self, other):
        diff = Configuration(None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None)

        if self.hostname != other.hostname:
            diff.__hostname = self.hostname

        if self.packs != other.packs:
            diff.__packs = self.packs

        if self.afe_id != other.afe_id:
            diff.__afe_id = self.afe_id

        if self.aws_api_auth != other.aws_api_auth:
            diff.__aws_api_auth = self.aws_api_auth

        if self.aws_client_auth != other.aws_client_auth:
            diff.__aws_client_auth = self.aws_client_auth

        if self.aws_group_config != other.aws_group_config:
            diff.__aws_group_config = self.aws_group_config

        if self.aws_project != other.aws_project:
            diff.__aws_project = self.aws_project

        if self.csv_logger_conf != other.csv_logger_conf:
            diff.__csv_logger_conf = self.csv_logger_conf

        if self.display_conf != other.display_conf:
            diff.__display_conf = self.display_conf

        if self.gas_baseline != other.gas_baseline:
            diff.__gas_baseline = self.gas_baseline

        if self.gas_model_conf != other.gas_model_conf:
            diff.__gas_model_conf = self.gas_model_conf

        if self.gps_conf != other.gps_conf:
            diff.__gps_conf = self.gps_conf

        if self.greengrass_identity != other.greengrass_identity:
            diff.__greengrass_identity = self.greengrass_identity

        if self.interface_conf != other.interface_conf:
            diff.__interface_conf = self.interface_conf

        if self.mpl115a2_calib != other.mpl115a2_calib:
            diff.__mpl115a2_calib = self.mpl115a2_calib

        if self.mpl115a2_conf != other.mpl115a2_conf:
            diff.__mpl115a2_conf = self.mpl115a2_conf

        if self.mqtt_conf != other.mqtt_conf:
            diff.__mqtt_conf = self.mqtt_conf

        if self.ndir_conf != other.ndir_conf:
            diff.__ndir_conf = self.ndir_conf

        if self.opc_conf != other.opc_conf:
            diff.__opc_conf = self.opc_conf

        if self.pmx_model_conf != other.pmx_model_conf:
            diff.__pmx_model_conf = self.pmx_model_conf

        if self.psu_conf != other.psu_conf:
            diff.__psu_conf = self.psu_conf

        if self.psu_version != other.psu_version:
            diff.__psu_version = self.psu_version

        if self.pt1000_calib != other.pt1000_calib:
            diff.__pt1000_calib = self.pt1000_calib

        if self.scd30_baseline != other.scd30_baseline:
            diff.__scd30_baseline = self.scd30_baseline

        if self.scd30_conf != other.scd30_conf:
            diff.__scd30_conf = self.scd30_conf

        if self.schedule != other.schedule:
            diff.__schedule = self.schedule

        if self.shared_secret != other.shared_secret:
            diff.__shared_secret = self.shared_secret

        if self.sht_conf != other.sht_conf:
            diff.__sht_conf = self.sht_conf

        if self.networks != other.networks:
            diff.__networks = self.networks

        if self.modem != other.modem:
            diff.__modem = self.modem

        if self.sim != other.sim:
            diff.__sim = self.sim

        if self.system_id != other.system_id:
            diff.__system_id = self.system_id

        if self.timezone_conf != other.timezone_conf:
            diff.__timezone_conf = self.timezone_conf

        return diff


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager):
        if self.hostname:
            raise ValueError('hostname may not be set')

        if self.packs:
            raise ValueError('packs may not be set')

        if self.afe_baseline:
            self.afe_baseline.save(manager)

        if self.afe_id:
            raise ValueError('afe_id may not be set')

        if self.aws_api_auth:
            self.aws_api_auth.save(manager)

        if self.aws_client_auth:
            self.aws_client_auth.save(manager)

        if self.aws_group_config:
            self.aws_group_config.save(manager)

        if self.aws_project:
            self.aws_project.save(manager)

        if self.csv_logger_conf:
            self.csv_logger_conf.save(manager)

        if self.display_conf:
            self.display_conf.save(manager)

        if self.gas_baseline:
            self.gas_baseline.save(manager)

        if self.gas_model_conf:
            self.gas_model_conf.save(manager)

        if self.gps_conf:
            self.gps_conf.save(manager)

        if self.greengrass_identity:
            self.greengrass_identity.save(manager)

        if self.interface_conf:
            self.interface_conf.save(manager)

        if self.mpl115a2_calib:
            self.mpl115a2_calib.save(manager)

        if self.mpl115a2_conf:
            self.mpl115a2_conf.save(manager)

        if self.mqtt_conf:
            self.mqtt_conf.save(manager)

        if self.ndir_conf:
            self.ndir_conf.save(manager)

        if self.opc_conf:
            self.opc_conf.save(manager)

        if self.pmx_model_conf:
            self.pmx_model_conf.save(manager)

        if self.psu_conf:
            self.psu_conf.save(manager)

        if self.psu_version:
            raise ValueError('psu_version may not be set')

        if self.pt1000_calib:
            self.pt1000_calib.save(manager)

        if self.scd30_baseline:
            self.scd30_baseline.save(manager)

        if self.scd30_conf:
            self.scd30_conf.save(manager)

        if self.schedule:
            self.schedule.save(manager)

        if self.shared_secret:
            self.shared_secret.save(manager)

        if self.sht_conf:
            self.sht_conf.save(manager)

        if self.networks:
            raise ValueError('networks may not be set')

        if self.modem:
            raise ValueError('modem may not be set')

        if self.sim:
            raise ValueError('sim may not be set')

        if self.system_id:
            self.system_id.save(manager)

        if self.timezone_conf:
            self.timezone_conf.save(manager)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['hostname'] = self.hostname
        jdict['packs'] = self.packs

        jdict['afe-baseline'] = self.afe_baseline
        jdict['afe-id'] = self.afe_id
        jdict['aws-api-auth'] = self.aws_api_auth
        jdict['aws-client-auth'] = self.aws_client_auth
        jdict['aws-group-config'] = self.aws_group_config
        jdict['aws-project'] = self.aws_project
        jdict['csv-logger-conf'] = self.csv_logger_conf
        jdict['display-conf'] = self.display_conf
        jdict['gas-baseline'] = self.gas_baseline
        jdict['gas-model-conf'] = self.gas_model_conf
        jdict['gps-conf'] = self.gps_conf
        jdict['interface-conf'] = self.interface_conf
        jdict['greengrass-identity'] = self.greengrass_identity
        jdict['mpl115a2-calib'] = self.mpl115a2_calib
        jdict['mpl115a2-conf'] = self.mpl115a2_conf
        jdict['mqtt-conf'] = self.mqtt_conf
        jdict['ndir-conf'] = self.ndir_conf
        jdict['opc-conf'] = self.opc_conf
        jdict['pmx-model-conf'] = self.pmx_model_conf
        jdict['psu-conf'] = self.psu_conf
        jdict['psu-version'] = self.psu_version
        jdict['pt1000-calib'] = self.pt1000_calib
        jdict['scd30-baseline'] = self.scd30_baseline
        jdict['scd30-conf'] = self.scd30_conf
        jdict['schedule'] = self.schedule
        jdict['shared-secret'] = self.shared_secret
        jdict['sht-conf'] = self.sht_conf
        jdict['sht-conf'] = self.sht_conf
        jdict['networks'] = self.networks
        jdict['modem'] = self.modem
        jdict['sim'] = self.sim
        jdict['system-id'] = self.system_id
        jdict['timezone-conf'] = self.timezone_conf

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def hostname(self):
        return self.__hostname


    @property
    def packs(self):
        return self.__packs


    @property
    def afe_baseline(self):
        return self.__afe_baseline


    @property
    def afe_id(self):
        return self.__afe_id


    @property
    def aws_api_auth(self):
        return self.__aws_api_auth


    @property
    def aws_client_auth(self):
        return self.__aws_client_auth


    @property
    def aws_group_config(self):
        return self.__aws_group_config


    @property
    def aws_project(self):
        return self.__aws_project


    @property
    def csv_logger_conf(self):
        return self.__csv_logger_conf


    @property
    def display_conf(self):
        return self.__display_conf


    @property
    def gas_baseline(self):
        return self.__gas_baseline


    @property
    def gas_model_conf(self):
        return self.__gas_model_conf


    @property
    def gps_conf(self):
        return self.__gps_conf


    @property
    def greengrass_identity(self):
        return self.__greengrass_identity


    @property
    def interface_conf(self):
        return self.__interface_conf


    @property
    def mpl115a2_calib(self):
        return self.__mpl115a2_calib


    @property
    def mpl115a2_conf(self):
        return self.__mpl115a2_conf


    @property
    def mqtt_conf(self):
        return self.__mqtt_conf


    @property
    def ndir_conf(self):
        return self.__ndir_conf


    @property
    def opc_conf(self):
        return self.__opc_conf


    @property
    def pmx_model_conf(self):
        return self.__pmx_model_conf


    @property
    def psu_conf(self):
        return self.__psu_conf


    @property
    def psu_version(self):
        return self.__psu_version


    @property
    def pt1000_calib(self):
        return self.__pt1000_calib


    @property
    def scd30_baseline(self):
        return self.__scd30_baseline


    @property
    def scd30_conf(self):
        return self.__scd30_conf


    @property
    def schedule(self):
        return self.__schedule


    @property
    def shared_secret(self):
        return self.__shared_secret


    @property
    def sht_conf(self):
        return self.__sht_conf


    @property
    def networks(self):
        return self.__networks


    @property
    def modem(self):
        return self.__modem


    @property
    def sim(self):
        return self.__sim


    @property
    def system_id(self):
        return self.__system_id


    @property
    def timezone_conf(self):
        return self.__timezone_conf


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Configuration:{hostname:%s, packs:%s, afe_baseline:%s, afe_id:%s, aws_api_auth:%s, " \
               "aws_client_auth:%s, aws_group_config:%s, aws_project:%s, csv_logger_conf:%s, display_conf:%s, " \
               "gas_baseline:%s, gas_model_conf:%s, gps_conf:%s, greengrass_identity:%s, interface_conf:%s, " \
               "mpl115a2_calib:%s, mpl115a2_conf:%s, mqtt_conf:%s, ndir_conf:%s, opc_conf:%s, " \
               "pmx_model_conf:%s, psu_conf:%s, psu_version:%s, pt1000_calib:%s, scd30_baseline:%s, " \
               "scd30_conf:%s, schedule:%s, shared_secret:%s, sht_conf:%s, networks:%s, modem:%s, " \
               "sim:%s, system_id:%s, timezone_conf:%s}" % \
               (self.hostname, self.packs, self.afe_baseline, self.afe_id, self.aws_api_auth,
                self.aws_client_auth, self.aws_group_config, self.aws_project, self.csv_logger_conf, self.display_conf,
                self.gas_baseline, self.gas_model_conf, self.gps_conf, self.greengrass_identity, self.interface_conf,
                self.mpl115a2_calib, self.mpl115a2_conf, self.mqtt_conf, self.ndir_conf, self.opc_conf,
                self.pmx_model_conf, self.psu_conf, self.psu_version, self.pt1000_calib, self.scd30_baseline,
                self.scd30_conf, self.schedule, self.shared_secret, self.sht_conf, self.networks, self.modem,
                self.sim, self.system_id, self.timezone_conf)
